/* -*- mode: c++; c-basic-offset: 4; -*- */
#include "gds_shmem.hh"
#include <cstdio>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <unistd.h>
#include <error.h>
#include <errno.h>
#include <new>

#define SHMPROT 0666
#if defined(sun)
#define SHMAT_MMU_FLG SHM_SHARE_MMU
#else
#define SHMAT_MMU_FLG 0
#endif

//======================================  Inline private flag manipulation 
inline void 
gds_shmem::clr_flag(enum flagbits flg) {
    _flagwd &= ~(1 << int(flg));    
}

//======================================  Construct shm descriptor
inline void 
gds_shmem::set_flag(enum flagbits flg) {
    _flagwd |= (1 << int(flg));
}

//======================================  Construct shm descriptor
gds_shmem::gds_shmem(void) 
    : _flagwd(0), _ID(0), _addr(0), _size(0), _protfl(SHMPROT)
{
}
 
//======================================  Destroy shm descriptor
gds_shmem::gds_shmem::~gds_shmem(void) {
    release();
}

//======================================  Attach a shm
bool 
gds_shmem::attach(int id) {
    //----------------------------------  Find an existing partition.
    if (!find(id, 0, 0)) return false;

    //----------------------------------  Attach it.
    return map();
}

//======================================  Create a shm
bool 
gds_shmem::create(int id, size_t minsiz) {
    //----------------------------------  Find the ID
    if (!find(id, minsiz, IPC_CREAT + _protfl)) return false;
    _size = minsiz;

    //----------------------------------  Map it into the memory space
    return map();
}

//======================================  Deaccess the partition.
bool
gds_shmem::deaccess(bool rmv) {
    if (!is_accessed()) return false;
    if (rmv) shmctl(_ID, IPC_RMID, 0);
    clr_flag(kAccessed);
    return true;
}

//======================================  Find an shm
bool 
gds_shmem::exists(int id, size_t minsize) {
    bool rv = find(id, minsize, 0);
    if (rv) release(false);
    return rv;
}

//======================================  Find an shm
bool 
gds_shmem::find(int id, size_t minsiz, int flag) {
    _error = 0;

    //----------------------------------  Make sure descriptor is available
    if (is_accessed()) return false;

    //----------------------------------  Get an ID
    _ID = shmget(id, minsiz, flag);
    if (_ID < 0) {
	_error = errno;
	// perror("gds_shmem::find - shmget");
	return false;
    }
    set_flag(kAccessed);
    return true;
}

//======================================  Lock the partition into memory
bool
gds_shmem::lock(bool flag) {
    bool rv = false;
    if (is_attached()) {
#if defined(sun)
	int saveuid = geteuid();
	if (saveuid) seteuid(0);
#endif
#if defined(P__DARWIN) || defined(P__WIN32)
	_error = EPERM;
#else
	if (geteuid()) {
	    return false;
	} 
	else if (flag) {
	    rv = !(shmctl(_ID, SHM_LOCK, 0) < 0);
	    if (rv) set_flag(kLocked);
	    else    _error = errno;
	}
	else {
	    rv = !(shmctl(_ID, SHM_UNLOCK, 0) < 0);
	    if (rv) clr_flag(kLocked);
	    else    _error = errno;
	}
#endif
#if defined(sun)
	if (saveuid) seteuid(saveuid);
#endif
    }
    return rv;
}

//======================================  Release the partition.
bool
gds_shmem::map(void) {
    //----------------------------------  Check state
    if (is_attached() || !is_accessed()) return false;

    //----------------------------------  Map it
    _addr = shmat(_ID, 0, SHMAT_MMU_FLG);
    if (long(_addr) == -1) {
	_error = errno;
	perror("gds_shmem::map - shmgat");
	return false;
    }
    set_flag(kAttached);
    return true;
}

//======================================  Release the partition.
void 
gds_shmem::release(bool remove) {
    //----------------------------------  Unlock locked segments
    if (is_locked()) {
	lock(false);
    }

    //----------------------------------  Unmap mapped segments
    if (is_attached()) {
	unmap();
    }

    //----------------------------------  Deaccess accessed segments
    if (is_accessed()) {
	deaccess(remove);
    }
}

//======================================  Release the partition.
void 
gds_shmem::set_mmd(void) {
    size_t* ptr = reinterpret_cast<size_t*>(_addr);
    *ptr++ = _size;
    *ptr++ = 2 * sizeof(size_t);
}

//======================================  Release the partition.
bool
gds_shmem::unmap(void) {
    if (!is_attached()) return false;
    if (shmdt(_addr) < 0) {
	perror("gds_shm::unmap - shmdt");
	return false;
    }
    clr_flag(kAttached);
    _addr = 0;
    return true;
}

void* 
operator new (size_t nb, gds_shmem& part) {
    size_t* av = reinterpret_cast<size_t*>(part.ref());
    size_t stoff = __sync_fetch_and_add(av+1, nb);
    if (stoff + nb > av[0]) {
	__sync_fetch_and_sub(av+1, nb);
	throw std::bad_alloc();
    }
    return reinterpret_cast<void*>(size_t(av) + stoff);
}

void* 
operator new[] (size_t nb, gds_shmem& part) {
    size_t* av = reinterpret_cast<size_t*>(part.ref());
    size_t stoff = __sync_fetch_and_add(av+1, nb);
    if (stoff + nb > av[0]) {
	__sync_fetch_and_sub(av+1, nb);
	throw std::bad_alloc();
    }
    return reinterpret_cast<void*>(size_t(av) + stoff);
}
