/* -*- mode: c++; c-basic-offset: 4; -*- */
//
//    Skeleton Frame processing routine Template
//
#include "PDnMon.hh"

#ifndef __CINT__

//-->  The next three lines are needed if you are going to generate triggers.
//     The text in PIDTITLE should describe the monitor function.
#define PIDCVSHDR "$Header: https://redoubt.ligo-wa.caltech.edu/svn/gds/trunk/Monitors/PDnMon/PDnMon.cc 6385 2010-12-10 23:00:42Z john.zweizig@LIGO.ORG $"
#define PIDTITLE  "Photodiode consistency check"
// #include "ProcIdent.hh"
#include "Dacc.hh"
#include "ParseLine.hh"
#include "DVector.hh"
#include "FilterDesign.hh"
#include "leqsolve.h"
#include <iostream>
#include <string>

using namespace std;

const Interval tSettle(2.0);
const long     maxAvg(100);

//======================================  Generate the main routine.
EXECDAT(PDnMon)
#endif               // !def(__CINT__)

//======================================  Skeleton object constructor.
PDnMon::PDnMon(int argc, const char *argv[]) 
    : DatEnv(argc, argv), maxFrame(999999), mStep(4.0), mOSC(mIn), 
      mAlarm("PDnMon")
{
    //----------------------------------  Define parameters
    mDict.addPar("MonName",     string("PDnMon"));
    mDict.addPar("OSCFile",     string(""));
    mDict.addPar("Stride",      double(4.0));
    mDict.addPar("SummaryTime", double(43200.0));

    //----------------------------------  Look for arguments
    string cfile = "PDnMon.cfg";

    bool syntax  = false;
    for (int i=1 ; i<argc ; i++) {
        string argi = argv[i];
        if (argi == "-conf") {
	    cfile = argv[++i];
 	} else if (isDatEnvArg(argv[i])) {
	    i++;
	} else {
	    syntax = true;
	    cerr << "Unrecognized argument: " << argi << endl;
	}
    }

    //----------------------------------  Bail out if command line not correct
    if (syntax) {
        cerr << "Command syntax:" << endl;
	cerr << argv[0] << " -conf <file> " << endl;
	finish();
	return;
    }

    //----------------------------------  Set up the trender
    mTrend.setName("PDnMon");    //  Set trend frame name
    mTrend.setFrameCount(1);          //  Set frames per file
    mTrend.setType(Trend::kMinute);   //  Set trend type (minute trends)
    mTrend.setAutoUpdate(false);      //  Avoid automatic screwup.

    //----------------------------------  Make sure there's a valid channel
    if (configure(cfile)) {
	cerr << "Unable to configure monitor" << endl;
	finish();
	return;
    }

    //----------------------------------  Set monitor name everywhere
    // char username[32];
    // cuserid(username);
    string pName = mDict.getString("MonName");
    mTrend.setName(pName.c_str()); //  Set trend name
    setServerName(pName.c_str());

    //----------------------------------  set the data accessor
    getDacc().setStride(mStep);
    getDacc().setIgnoreMissingChannel(true);
    mOSC.setStride(mStep);
    setStrideAlignment(int(mStep), 0.0);

    //----------------------------------  set the alarm client
    mAlarm.defineAlarm("GainLow", mStep, 3, "Gain for PD: $1 low: $2");
}

//======================================  Skeleton object destructor.
PDnMon::~PDnMon() 
{
    cout << "PDnMon is finished" << endl;
}

//======================================  Frame processing function.
bool
PDnMon::configure(const std::string& cfile) {

    //----------------------------------  Open the configuration file
    ParseLine pl(cfile.c_str());
    if (!pl.isOpen()) {
	cerr << "Unable to open file: " << cfile << endl;
	return true;
    }
    if (Debug()) pl.setLog(cerr);

    //----------------------------------  Snort the command lines.
    while (pl.getLine() >= 0) {
	int nArg = pl.getCount();
	if (!nArg) continue;
	string cmd = pl[0];

	//------------------------------  Group command
	if (cmd == "Group") {
	    mList.push_back(group(pl[1], mIn, mTrend));
	    double sRate = 16384.0;
	    int    debugLevel = Debug();
	    for (int i=2; i<nArg; ++i) {
		string argi = pl[i];
		if (argi == "-debug") {
		    if (i+1 < nArg) debugLevel = strtol(pl[++i], 0, 0);
		    else            debugLevel = 1;
		} else if (argi == "-filter") {
		    FilterDesign fd(pl[++i], sRate);
		    mList.back().setFilter(fd.get());
		} else if (argi == "-rate") {
		    sRate = pl.getDouble(++i);
		} else if (argi == "-thresh") {
		    mList.back().setThreshold(pl.getDouble(++i));
		} else if (argi == "-while") {
		    mList.back().setCondition(pl[++i]);
		} else {
		    cerr << "Unrecognized Group keyword: " << argi << endl;
		    return true;
		}
	    }
	    mList.back().setDebug(debugLevel);	    

	//------------------------------  Channel command
	} else if (cmd == "Channel") {
	    subchan sb(pl[1], mTrend);
	    for (int i=2; i<nArg; ++i) {
		string argi = pl[i];
		if (argi == "-quad") {
		    sb.setQuadName(pl[++i]);
		} else {
		    cerr << "Undefined keyword: " << argi << endl;
		}
	    }
	    mList.back().addSubChan(sb);

        //------------------------------  Parse parameter line
        } else if (cmd == "Parameter") {
            for (int i=1 ; i<nArg ; i+=2) {
                if (!mDict.exists(pl[i])) {
                    cerr << "No such parameter: " << pl[i] << endl;
                    break;
                } else {
                    if (Debug()) cout << "Set parameter: " << pl[i] 
                                      << " to: " << pl[i+1] << endl;
                    mDict.setPar(pl[i], pl[i+1]);
                }
            }

        //------------------------------  Other/error
	} else {
	    cerr << "Unrecognized command name: " << cmd << endl;
	    return true;
	}
    }

    //---------------------------------- Get the OSC condition definitions
    string oscfile = mDict.getString("OSCFile");
    if (!oscfile.empty()) {
	mOSC.readConfig(oscfile.c_str());
    }
    mStep = mDict.getDouble("Stride");

    //---------------------------------- Set up the dmtviewer objects
    if (Debug()) cerr << "Setting up dmtviewer objects for groups:" << endl;
    for (group_iter i=mList.begin(); i != mList.end(); ++i) {
	if (Debug()) cerr << "Group: " << i->getName() << endl;
	i->setViewer(*this, Interval(mDict.getDouble("SummaryTime")));
    }
    if (Debug()) cerr << endl;
    return false;
}

//======================================  Frame processing function.
void
PDnMon::ProcessData(void) {
    //----------------------------------  Loop over channel groups
    for (group_iter i=mList.begin(); i != mList.end(); ++i) {
	const char* oscCond = i->getCondition();
	if (oscCond && *oscCond && mOSC.satisfied(oscCond)) {
	    i->calculate();
	    i->alarm(mAlarm);
	}
    }
    mTrend.Update(mIn.getCurrentTime());
    if (Now().getS()%60 < mStep) printStat();
}

//======================================  Frame processing function.
void
PDnMon::printStat(void) const {
    //----------------------------------  Open the statistics file.
    const char* html_dir=getenv("DMTHTMLOUT");
    std::string file;
    if (html_dir) {
	file  = html_dir;
	file += "/";
    }
    file += "PDnMon_stats.txt";
    ofstream out(file.c_str());
    if (!out.is_open()) {
	cerr << "Unable to open statistics file: " << file << endl;
    }

    //----------------------------------  Header text
    char datestr[64];
    out << "Statistics at: " 
	<< TimeStr(Now(), datestr, "%m/%d/%Y %H:%N:%S %Z") 
	<< endl;

    //----------------------------------  Loop over channel groups
    for (const_group_iter i=mList.begin(); i != mList.end(); ++i) {
	i->printStat(out);
    }
}

//======================================  Handle Messages
void
PDnMon::Attention(void) {
    MonServer::Attention();
}

//======================================  group constructor
PDnMon::group::group(const std::string& name, Dacc& dacc, Trend& tr) 
    : _name(name), _dacc(dacc), _debug(0), _trend(tr), _thresh(3.0), 
      _scale_alarm(0.9), _bias_alarm(0.5), _current(0), 
      _histo(name.c_str(), 100, 0, 10.0)
{
    _troot = name;
    for (string::size_type i=_troot.find("-"); i!=string::npos; ) {
	_troot[i] = '_';
	i = _troot.find("-");
    }
    string::size_type inx = _troot.find(":");
    if (inx != string::npos) {
	_troot.insert(inx+1, "DMT-PDNM_");
    }
    _trend.addChannel((_troot+"_Chi_Avg").c_str());
    _trend.addChannel((_troot+"_Glitch_Rate").c_str());
}

//======================================  group destructor
PDnMon::group::~group(void) {
}

//======================================  Add a sub-channel
void 
PDnMon::group::addSubChan(const std::string& name) {
    addSubChan(subchan(name, _trend));
}

//======================================  Add a sub-channel
void 
PDnMon::group::addSubChan(const subchan& sb) {
    _list.push_back(sb);
    _dacc.addChannel(_list.back().getName());
    if (!sb.refQuadName().empty()) {
	_dacc.addChannel(sb.refQuadName().c_str());
    }
    if (!_pipe.null()) _list.back().setFilter(*_pipe);
    _list.back().setDebug(_debug);
}


//======================================  Generate alarms as appropriate
void 
PDnMon::group::alarm(AlarmClient& ac) {
    AlarmData ad("PDnMon", "GainLow", 0, 3);
    AlarmHandle ah;

    //----------------------------------  Throw an alarm based on linear xform
    for (subchan_iter i=_list.begin(); i != _list.end(); ++i) {
	double gain = i->getScale();
	if (i->getScale() < _scale_alarm) {
	    ostringstream pars;
	    pars << i->getName() << " " << gain;
	    ad.setParameters(pars.str());
	    ac.setAlarm(ad, ah);
	}
    }    
}

//======================================  group constructor
void 
PDnMon::group::calculate(void) {

    //----------------------------------  Check for data dropouts reset if so
    if (_current != _dacc.getFillTime()) {
	for (subchan_iter i=_list.begin(); i != _list.end(); ++i) {
	    i->reset();
	}
	_current  = _dacc.getFillTime();
	_segstart = _current;
	_weight   = 0.0;
	_avgcnt   = 0;
    }
    Time tStart = _current;

    //----------------------------------  Read in all sub-channel data
    for (subchan_iter i=_list.begin(); i != _list.end(); ++i) {
	i->readChan(_dacc);
    }
    _current = _dacc.getCurrentTime();
    if (tStart < _segstart + tSettle) return;

    //----------------------------------  calculate the average signal.
    int nSubC = 0;
    for (subchan_iter i=_list.begin(); i != _list.end(); ++i) {
	if (i == _list.begin()) _avg  = i->getData();
	else                    _avg += i->getData();
	nSubC++;
    }
    _avg /= double(nSubC);
    
    //----------------------------------  calculate the sample-by-sample Chi^2
    TSeries chi;
    for (subchan_iter i=_list.begin(); i != _list.end(); ++i) {
	i->scale(_avg);
	if (i == _list.begin()) {
	    chi.copy(i->sqoff(_avg));
	    chi.Convert(DVector::t_double);
	} else {
	    chi += i->sqoff(_avg);
	}
    }

    //----------------------------------  calculate a running weighted average
    long   N = chi.getNSample();
    double chiAvg = chi.getAverage();
    if (_avgcnt < 10 || chiAvg < 5*_weight) {
	double nAvg   = (_avgcnt++ < maxAvg) ? _avgcnt : maxAvg;
	_weight = (_weight * (nAvg - 1.0) + chiAvg) / nAvg;
    }

    //----------------------------------  record statistics
    if (_weight != 0.0) {
	Interval dT = chi.getInterval();
	float nOver = chi.getNGreater(_thresh*_weight)/double(dT);
	float wtAvg = chiAvg / _weight;
	_rhist.fixedAppend(tStart, dT, &nOver);
	_chist.fixedAppend(tStart, dT, &wtAvg);
	_histo.Clear();
	chi *= double(nSubC-1)*_weight;
	_histo.FillN(N, reinterpret_cast<const double*>(chi.refData()));
	_trend.trendData((_troot+"_Chi_Avg").c_str(), tStart, wtAvg);
	_trend.trendData((_troot+"_Glitch_Rate").c_str(), tStart, nOver);
    }
}

//======================================  Print group statistics
void 
PDnMon::group::printStat(ostream& out) const {
    out << "Group: " << _name << endl;
    out << "  Weight:       " << _weight << endl;
    out << "  Weight count: " << _avgcnt << endl;
    
    for (const_subchan_iter i=_list.begin(); i != _list.end(); ++i) {
	i->printStat(out);
    }
}

//======================================  Set group filter.
void 
PDnMon::group::setFilter(const Pipe& pipe) {
    _pipe.set(pipe);
    _current = Time(0);
}

//======================================  Set the threshold.
void 
PDnMon::group::setThreshold(double thresh) {
    _thresh = thresh;
}

//======================================  Define group viewer objects.
void 
PDnMon::group::setViewer(MonServer& ms, Interval histl) {

    //----------------------------------  Average Chi^2 history
    _chist.setMaxLen(histl);
    string objName = _name + "_Chi_Avg";
    if (_debug) cerr << "   Serve: " << objName << endl; 
    ms.serveData(objName.c_str(), &_chist);

    //----------------------------------  Glitch rate history
    _rhist.setMaxLen(histl);
    objName = _name + "_Glitch_Rate";
    if (_debug) cerr << "   Serve: " << objName << endl; 
    ms.serveData(objName.c_str(), &_rhist);

    //----------------------------------  Histogram Chi^2 terms
    objName = _name + "_Chi2_Histo";
    if (_debug) cerr << "   Serve: " << objName << endl; 
    ms.serveData(objName.c_str(), &_histo);

    //----------------------------------  Average signal TSeries.
    objName = _name + "_Average";
    if (_debug) cerr << "   Serve: " << objName << endl; 
    ms.serveData(objName.c_str(), &_avg);

    //----------------------------------  Channel objects.
    for (subchan_iter i=_list.begin(); i != _list.end(); ++i) {
	i->setViewer(ms, histl);
    }
}

//======================================  Sub-channel constructor
PDnMon::subchan::subchan(const std::string& name, Trend& tr)
    : _name(name), _debug(0), _trend(tr), _off(0.0), _scale(1.0),
      _phase(0)
{
    _troot = name;
    for (string::size_type i=_troot.find("-"); i!=string::npos; ) {
	_troot[i] = '_';
	i = _troot.find("-");
    }
    string::size_type inx = _troot.find(":");
    if (inx != string::npos) {
	_troot.insert(inx+1, "DMT-PDNM_");
    }
    _trend.addChannel((_troot+"_Scale").c_str());
    _trend.addChannel((_troot+"_Bias").c_str());
}

//======================================  Sub-channel destructor
PDnMon::subchan::~subchan(void) {
}

//======================================  Get scaled/phase correct data
TSeries 
PDnMon::subchan::getScaledData(void) const {
    _scratch.copy(_raw);
    _scratch -= _off;
    _scratch *= 1.0/_scale;
    if (_phase != 0) {
	TSeries tmp(_quad);
	tmp *= _phase;
	_scratch -= tmp;
    }
    return _scratch;
}

//======================================  Print channel statistics
void 
PDnMon::subchan::printStat(ostream& out) const {
    out << "  Channel: " << _name << endl;
    out << "    Scale:       " << _scale << endl;
    out << "    Offset:      " << _off << endl;
    out << "    Phase:       " << _phase << endl;
}

//======================================  Get channel data
void 
PDnMon::subchan::readChan(const Dacc& dacc) {
    const TSeries* r = dacc.refData(_name.c_str());
    if (r) {
	if (_pipe.null()) _raw = *r;
	else              _raw = _pipe(*r);
    } else {
	cerr << "Can't find TSeries for channel: " << _name << endl;
	dacc.list(cerr);
	throw runtime_error("Error processing channel");
    }
    if (!_quadname.empty()) {
	const TSeries* q = dacc.refData(_quadname.c_str());
	if (q) {
	    if (_qpipe.null()) _quad = *q;
	    else               _quad = _qpipe(*q);
	}
    }
}

//======================================  Calculate te scale factors.
void 
PDnMon::subchan::reset(void) {
    _pipe.reset();
    _qpipe.reset();
}

//======================================  Calculate the scale factors.
inline double
dig5(double x) {
    return long((x+0.00005)*10000.0)/10000.0;
}

//======================================  Calculate the scale factors.
void 
PDnMon::subchan::scale(const TSeries& avg) {
    double Aww = avg.getNSample();
    _scale = 1.0;
    _off   = 0.0;
    _phase = 0.0;
    if (Aww != double(_raw.getNSample())) {
	cerr << "Average length: " << Aww << "  Channel length: " 
	     << _raw.getNSample() << endl;
	throw runtime_error("Average vector length differs from channel");
    } else if (Aww != 0) {
	if (_quadname.empty() || _quad.isEmpty()) {
	    double A[4], V[2], X[2];
	    A[0] = Aww;
	    A[1] = avg.getAverage()*Aww;
	    A[2] = A[1];
	    A[3] = avg*avg;

	    V[0] = _raw.getAverage()*Aww;
	    V[1] = _raw*avg;

	    if (!solve22(A, V, X)) {
		_off   = X[0];
		_scale = X[1];
	    }
	} else {
	    double A[9], V[3], X[3];
	    A[0] = Aww;
	    A[1] = avg.getAverage()*Aww;
	    A[3] = A[1];
	    A[2] = _quad.getAverage()*Aww;
	    A[6] = A[2];
	    A[4] = avg*avg;
	    A[5] = _quad*avg;
	    A[7] = A[5];
	    A[8] = _quad*_quad;

	    V[0] = _raw.getAverage()*Aww;
	    V[1] = _raw*avg;
	    V[2] = _raw*_quad;

	    if (!solve33(A, V, X)) {
		_off   = X[0];
		_scale = X[1];
		_phase = X[2];
	    }
	}
    }

    //----------------------------------  record statistics
    Time tStart = avg.getStartTime();
    if (_debug) {
	cerr << "PDnMon: " << _name << " transform at: " << tStart.getS()
	     << " scale: " << dig5(_scale) << " bias: " << dig5(_off) << endl;
    }
    Interval dT = avg.getInterval();
    float temp = _scale;
    _shist.fixedAppend(tStart, dT, &temp);
    temp = _off;
    _bhist.fixedAppend(tStart, dT, &temp);
    _trend.trendData((_troot+"_Scale").c_str(), tStart, _scale);
    _trend.trendData((_troot+"_Bias").c_str(), tStart, _off);
}

//======================================  group constructor
void 
PDnMon::subchan::setFilter(const Pipe& pipe) {
    _pipe.set(pipe);
    _qpipe.set(pipe);
}

//======================================  group constructor
void 
PDnMon::subchan::setQuadName(const string& name) {
    _quadname = name;
}

//======================================  group constructor
void 
PDnMon::subchan::setViewer(MonServer& ms, Interval histl) {
    _bhist.setMaxLen(histl);
    string objName = _name + "_Bias";
    if (_debug) cerr << "   Serve: " << objName << endl; 
    ms.serveData(objName.c_str(),  &_bhist);
    _shist.setMaxLen(histl);
    objName = _name + "_Scale";
    if (_debug) cerr << "   Serve: " << objName << endl; 
    ms.serveData(objName.c_str(), &_shist);
    objName = _name + "_Terms";
    if (_debug) cerr << "   Serve: " << objName << endl; 
    ms.serveData(objName.c_str(), &_terms);
}

//======================================  Sub-channel class
TSeries 
PDnMon::subchan::sqoff(const TSeries& avg) {
    _terms = getScaledData();
    _terms -= avg;
    _scratch.copy(_terms);
    *(_scratch.refDVect()) *= *(_terms.refDVect());
    return _scratch;
}
